/**
 * Downloader.java
 * Jeff Ondich, 22 May 2014
 *
 * When Downloader's download method is invoked, the Downloader spawns a
 * new thread, tries to download the file to the specified destination file,
 * and posts Message objects to the caller's Handler.
 *
 * Here are Message.what values sent by Downloader, along with the values of
 * Message.arg1, arg2, and obj for each type of message.
 *
 * HANDLE_SET_PROGRESS_MAX: arg1 is the new maximum value.
 * HANDLE_PROGRESS: arg1 is the new progress value.
 * HANDLE_SUCCESS: obj is the File object of the destination file.
 * HANDLE_ERROR: obj is the thrown Exception object
 * HANDLE_TIMEOUT: no parameters
 */

package com.jeffondich.threading.app;

import android.os.Handler;
import android.os.Message;
import android.util.Log;

import java.io.BufferedInputStream;
import java.io.File;
import java.io.FileOutputStream;
import java.io.InputStream;
import java.io.OutputStream;
import java.net.URL;
import java.net.URLConnection;
import java.util.HashMap;

public class Downloader {
    private int progressBarMax;
    private int progressBarProgress;

    private String sourceURL;
    private File destinationFile;
    private HashMap<String, String> requestHeaders;

    private Handler downloaderUpdatesHandler;

    private boolean timeoutHasOccurred;
    private int timeoutInMillis;
    private Handler timeoutHandler;
    private Runnable timeoutRunnable;

    public static final int HANDLE_SET_PROGRESS_MAX = 0x7162;
    public static final int HANDLE_PROGRESS = 0x7163;
    public static final int HANDLE_SUCCESS = 0x7164;
    public static final int HANDLE_ERROR = 0x7165;
    public static final int HANDLE_TIMEOUT = 0x7166;

    public Downloader() {
        this.timeoutHandler = new Handler();

        this.timeoutRunnable = new Runnable() {
            public void run() {
                timeoutHasOccurred = true;
            }
        };
    }

    public void download(String sourceURL,
                         File destinationFile,
                         int timeoutInSeconds,
                         HashMap<String, String> requestHeaders,
                         Handler handler)
    {
        this.sourceURL = sourceURL;
        this.destinationFile = destinationFile;
        this.requestHeaders = requestHeaders;
        this.timeoutInMillis = 1000 * timeoutInSeconds;
        this.downloaderUpdatesHandler = handler;

        Thread thread = new Thread(new DownloadRunnable());
        thread.setDaemon(true);

        this.timeoutHandler.postDelayed(this.timeoutRunnable, this.timeoutInMillis);

        thread.start();
    }

    private void publishProgressMax() {
        Message message = new Message();
        message.arg1 = progressBarMax;
        message.what = Downloader.HANDLE_SET_PROGRESS_MAX;
        this.downloaderUpdatesHandler.sendMessage(message);
    }

    private void publishProgress() {
        Message message = new Message();
        message.arg1 = progressBarProgress;
        message.what = Downloader.HANDLE_PROGRESS;
        this.downloaderUpdatesHandler.sendMessage(message);
        Log.d("Progress [Downloader]", String.format("(%d, %d)", progressBarProgress, progressBarMax));
    }

    private void publishTimeout() {
        Message message = new Message();
        message.what = Downloader.HANDLE_TIMEOUT;
        this.downloaderUpdatesHandler.sendMessage(message);
    }

    private void publishError(Exception e) {
        Message message = new Message();
        message.obj = e;
        message.what = Downloader.HANDLE_ERROR;
        this.downloaderUpdatesHandler.sendMessage(message);
    }

    private void publishSuccess() {
        Message message = new Message();
        message.what = Downloader.HANDLE_SUCCESS;
        message.obj = this.destinationFile;
        this.downloaderUpdatesHandler.sendMessage(message);
    }

    private class DownloadRunnable implements Runnable {
        public void run() {
            try {
                URL url = new URL(sourceURL);
                URLConnection connection = url.openConnection();
                for (String key : requestHeaders.keySet()) {
                    connection.setRequestProperty(key, requestHeaders.get(key));
                }
                connection.connect();
                int progressFileLength = connection.getContentLength();
                progressBarMax = (progressFileLength > 0 ? progressFileLength : 100);
                progressBarProgress = 0;
                publishProgressMax();

                InputStream inputStream = new BufferedInputStream(connection.getInputStream(), 8192);
                OutputStream outputStream = new FileOutputStream(destinationFile);

                int byteCount = 0;
                byte data[] = new byte[1024];
                while (!timeoutHasOccurred && (byteCount = inputStream.read(data)) != -1) {
                    outputStream.write(data, 0, byteCount);

                    timeoutHandler.removeCallbacks(timeoutRunnable);
                    timeoutHandler.postDelayed(timeoutRunnable, timeoutInMillis);

                    progressBarProgress += byteCount;
                    publishProgress();
                }

                outputStream.flush();
                outputStream.close();
                inputStream.close();

                if (timeoutHasOccurred) {
                    publishTimeout();
                } else {
                    publishSuccess();
                }

            } catch (Exception e) {
                publishError(e);
            }
        }
    }
}
